const
  MAX_RECUR_LEVEL = 50;
type
  TparserIdsStack = array [1..MAX_RECUR_LEVEL] of string;
  TparserException = class(Exception);

procedure applyMacrosAndSymbols(var txt:string; cb:TmacroCB; cbData:pointer; table:TStringDynArray;
  recurLevel:integer; var idsStack:TparserIdsStack);
const
  OPEN = '{{';
  CLOSE = '}}';
  SEP = '|';
  OPEN_ID = 0;
  CLOSE_ID = 1;
  SEP_ID = 2;
  QUOTE_ON_ID = 3;
  QUOTE_OFF_ID = 4;
  QUOTE = 'quote';
  MARKER_TAG: array [OPEN_ID..QUOTE_OFF_ID] of string = (OPEN, CLOSE, SEP, OPEN+QUOTE+' ', CLOSE);
type
  Tblock = record
    open, close: integer;
    seps:array of integer;
    quoting: boolean;
    end;
var
  Nblocks: dword;
  markers: array of dword;
  blocks: array of Tblock;

  function alreadyRecurredOn(s:string):boolean;
  var
    i: integer;
  begin
  result:=TRUE;
  for i:=recurLevel downto 1 do
    if sameText(s, idsStack[i]) then exit;
  result:=FALSE;
  end; // alreadyRecurredOn

  procedure handleSymbols();
  var
    b, e, l,p : integer;
    s: string;
  begin
  e:=0;
  l:=length(txt);
  while e < l do
    begin
    // search for next symbol
    b:=posEx('%',txt,e+1);
    if b = 0 then break;
    e:=b+1;
    while (e < l) and (txt[e] in ['a'..'z','A'..'Z','-','_']) do
      inc(e);
    if txt[e] <> '%' then continue;
    if e = b+1 then
      begin    // we don't accept %% as a symbol
      e:=b-1;  // so, restart parsing from the second %
      continue;
      end;
    // found!
    s:=substr(txt,b,e);
    if alreadyRecurredOn(s) then continue; // the user probably didn't meant to create an infinite loop
    // search for the symbol in the translation table
    p:=length(table)-2;
    if odd(p) then dec(p); // ensure the table has a legal length
    while (p >= 0) and not sameText(table[p], s) do
      dec(p,2);
    if p < 0 then continue;
    idsStack[recurLevel]:=s; // keep track of what we recur on
    // apply translation, and eventually recur
    try
      s:=table[p+1];
      if s > '' then
        applyMacrosAndSymbols(s, cb, cbData, table, recurLevel, idsStack);
    finally idsStack[recurLevel]:='' end;
    inc(e, replace(txt, s, b, e));
    l:=length(txt);
    end;
  end; // handleSymbols

  procedure collectMarkers();
  var
    i, j, n, m: integer;
    quoting: boolean;

    procedure addMarker(m:integer);
    begin
    markers[n]:=dword(i)+dword(m) shl 24;  // use the high byte to encode the marker
    inc(n);
    inc(i, length(MARKER_TAG[m])-1);
    end; // addMarker

  begin
  setLength(markers, length(txt) div 3); // it will never need more than that
  n:=0;
  i:=1;
  quoting:=FALSE;
  while i <= length(txt)-length(OPEN)+1 do
    begin
    for m:=OPEN_ID to SEP_ID do
      if strAt(txt, MARKER_TAG[m], i) then
        begin
        if quoting then
          if (m <> CLOSE_ID) or not strAt(txt, '/'+QUOTE, i-length(QUOTE)-1) then
            break
          else
            begin
            addMarker(QUOTE_OFF_ID);
            quoting:=FALSE;
            break;
            end;
        if m = OPEN_ID then
          begin
          j:=i+length(OPEN);
          if (j > length(txt)) or (txt[j] in [' ',#13,#10,#0]) then break // blocks must begin with macronames
          else if strAt(txt, QUOTE+' ', j) then
            begin
            addMarker(QUOTE_ON_ID);
            quoting:=TRUE;
            break;
            end;
          end;
        addMarker(m);
        break; // we found the marker, don't test others
        end;
    inc(i);
    end;
  setLength(markers, n); // be real
  end; // collectMarkers

  procedure checkMarkers();
  var
    i, n, open, ofs: integer;
  begin
  // check correctness of markers' order. we set to zero invalid ones.
  open:=0;
  Nblocks:=0;
  n:=length(markers);
  for i:=0 to n-1 do
    case markers[i] shr 24 of
      OPEN_ID:
        begin
        inc(open);
        inc(Nblocks);
        end;
      SEP_ID:
        if open <= 0 then
          markers[i]:=0;
      CLOSE_ID:
        if open > 0 then
          dec(open)
        else
          markers[i]:=0;
      QUOTE_ON_ID: inc(Nblocks);
      end;
  if open > 0 then raise TparserException.create(format('you left %d brakets open',[open]));
  // remove zeroes
  i:=0;
  ofs:=0;
  while i+ofs < integer(n) do
    begin
    open:=markers[i];
    if open = 0 then inc(ofs);
    if ofs > 0 then markers[i]:=markers[i+ofs];
    if open > 0 then inc(i);
    end;
  setLength(markers, integer(n)-ofs);
  end; // checkMarkers

  procedure markersToBlocks(); // organize MARKER_TAG in blocks
  var
    i, n, p, sp, bp: integer;
    stack: array of integer;
  begin
  setLength(blocks, Nblocks);
  setLength(stack, Nblocks);
  n:=0; // number of markers
  sp:=-1; // stack pointer
  bp:=-1; // block pointer
  for i:=0 to length(markers)-1 do
    begin
    p:=markers[i] and $FFFFFF;
    case markers[i] shr 24 of
      OPEN_ID:
        begin
        inc(sp);
        stack[sp]:=n;
        bp:=n;
        blocks[bp].open:=p;
        inc(n);
        end;
      CLOSE_ID:
        begin
        blocks[bp].close:=p;
        dec(sp);
        bp:=stack[sp];
        end;
      SEP_ID:
        if sp >= 0 then // if this marker is found outside macros then it's simple text
          with blocks[bp] do
            begin
            setLength(seps, length(seps)+1);
            seps[length(seps)-1]:=p;
            end;
      QUOTE_ON_ID:
        begin
        blocks[n].open:=p;
        blocks[n].quoting:=TRUE;
        end;
      QUOTE_OFF_ID:
        begin
        blocks[n].close:=p;
        inc(n);
        end;
      end;
    end;
  end; // markersToBlocks

  procedure sortBlocks();
  var
    i, k: integer;
  begin
  if Nblocks > 1 then
    for i:=0 to Nblocks-2 do
      for k:=i+1 to Nblocks-1 do
        swapMem(blocks[i], blocks[k], sizeOf(blocks[0]), blocks[i].close > blocks[k].close);
  end; // sortBlocks

  procedure expandMacros();
  var
    sp, i, j, o, d: integer;
    s, name, fullMacro: string;
    blk: ^Tblock;
    pars: Tstrings;
  begin
  pars:=TstringList.create;
  try
    for i:=0 to Nblocks-1 do
      begin
      blk:=@blocks[i];
      if blk.quoting then continue;
      fullMacro:=substr(txt, blk.open, blk.close+length(CLOSE)-1 );
      if alreadyRecurredOn(fullMacro) then continue; // the user probably didn't meant to create an infinite loop
      // extract macro name
      if blk.seps = NIL then j:=blk.close
      else j:=blk.seps[0];
      name:=substr(txt, blk.open+length(OPEN), j-1);

      pars.clear();
      // extract first n-1 parameters
      if length(blk.seps) > 1 then
        for j:=0 to length(blk.seps)-2 do
          pars.add(substr(txt, blk.seps[j]+length(SEP), blk.seps[j+1]-1));
      if length(blk.seps) > 0 then
        begin
        // extract last parameter
        s:=substr(txt, blk.seps[length(blk.seps)-1]+length(SEP), blk.close-1);
        if ansiEndsText('/'+name, s) then
          setLength(s, length(s)-length(name)-1);
        pars.add(s);
        end;
      // do the call, recur, and replace with the result
      s:=cb(name, pars, cbData);
      idsStack[recurLevel]:=fullmacro; // keep track of what we recur on
      try applyMacrosAndSymbols(s, cb, cbData, table, recurLevel, idsStack)
      finally idsStack[recurLevel]:='' end;
      d:=replace(txt, s, blk.open, blk.close+length(CLOSE)-1 );
      o:=blk.open;
      //** i think it can be optimized: remove the for-loop, and keep 2 ints (from,ofs) reporting the 'ofs' for every index > 'from' 
      // update offsets
      for j:=i+1 to Nblocks-1 do
        begin
        blk:=@blocks[j];
        if blk.open > o then inc(blk.open, d);
        if blk.close > o then inc(blk.close, d);
        for sp:=0 to length(blk.seps)-1 do
          if blk.seps[sp] > o then inc(blk.seps[sp], d);
        end;
      end;
  finally pars.free end;
  end; // expandMacros

  procedure handleMacros();
  begin
  collectMarkers();
  if markers = NIL then exit; // no MARKER_TAG, no macros
  checkMarkers();
  markersToBlocks();
  sortBlocks();
  expandMacros();
  end; // handleMacros
  
begin
if recurLevel > MAX_RECUR_LEVEL then exit;
inc(recurLevel);
handleSymbols();
handleMacros();
txt:=ansiReplaceText(txt, OPEN+QUOTE, '');
txt:=ansiReplaceText(txt, '/'+QUOTE+CLOSE, '');
end; //applyMacrosAndSymbols

function tryApplyMacrosAndSymbols(var txt:string; cb:TmacroCB; cbData:pointer; table:TStringDynArray):boolean;
var
  idsStack: TparserIdsStack;
begin
result:=FALSE;
try
  applyMacrosAndSymbols(txt, cb, cbData, table, 0, idsStack);
  result:=TRUE;
except
  on TparserException do; // quit silently
  on e:Exception do msgDlg(e.Message, MB_ICONERROR);
  end
end; // tryApplyMacrosAndSymbols

